from contextlib import closing
from typing import Any, ContextManager, cast
from uuid import uuid1
from warnings import warn

from docker import DockerClient

try:
    from kafka import KafkaConsumer, KafkaProducer
    from kafka.errors import KafkaError
except ImportError:
    KafkaConsumer = KafkaProducer = None
    # python3.12 uses confluent_kafka
    from confluent_kafka import Consumer as ConfluentConsumer
    from confluent_kafka.error import KafkaError

from yellowbox.containers import SafeContainerCreator, get_ports
from yellowbox.networks import anonymous_network
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, RunMixin, SingleEndpointService
from yellowbox.utils import DOCKER_EXPOSE_HOST, get_free_port

__all__ = ["KafkaService"]

warn(
    "bitnami is gone! use `yellowbox-kraft` instead",
    DeprecationWarning,
    stacklevel=1,
)


class KafkaService(SingleEndpointService, RunMixin, AsyncRunMixin):
    def __init__(
        self,
        docker_client: DockerClient,
        tag_or_images: str | tuple[str, str] = "latest",
        inner_port=0,
        outer_port=0,
        bitnami_debug: bool = False,
        **kwargs,
    ):
        self.inner_port = inner_port or get_free_port()
        self.outer_port = outer_port or get_free_port()
        if isinstance(tag_or_images, str):
            zookeeper_image = f"bitnami/zookeeper:{tag_or_images}"
            broker_image = f"bitnami/kafka:{tag_or_images}"
        else:
            zookeeper_image, broker_image = tag_or_images

        # broker must have a known alias at creation time
        self.static_broker_alias = f"broker-{uuid1()}"

        creator = SafeContainerCreator(docker_client)

        extra_bitnami_env = {}
        if bitnami_debug:
            extra_bitnami_env["BITNAMI_DEBUG"] = "true"

        self.zookeeper = creator.create_and_pull(
            zookeeper_image,
            detach=True,
            publish_all_ports=True,
            environment={
                "ZOOKEEPER_CLIENT_PORT": "2181",
                "ZOOKEEPER_TICK_TIME": "2000",
                "ALLOW_ANONYMOUS_LOGIN": "yes",
                **extra_bitnami_env,
            },
        )

        self.broker = creator.create_and_pull(
            broker_image,
            ports={
                str(self.outer_port): ("0.0.0.0", self.outer_port),
                str(self.inner_port): ("0.0.0.0", self.inner_port),
            },
            publish_all_ports=True,
            detach=True,
            environment={
                "KAFKA_CFG_ADVERTISED_LISTENERS": f"INNER://{self.static_broker_alias}:{self.inner_port},"
                f"OUTER://localhost:{self.outer_port}",
                "KAFKA_CFG_ZOOKEEPER_CONNECT": "zk/2181",
                "ALLOW_PLAINTEXT_LISTENER": "yes",
                "KAFKA_CFG_LISTENER_SECURITY_PROTOCOL_MAP": "INNER:PLAINTEXT,OUTER:PLAINTEXT",
                "KAFKA_INTER_BROKER_LISTENER_NAME": "INNER",
                "KAFKA_CFG_LISTENERS": f"INNER://:{self.inner_port},OUTER://:{self.outer_port}",
                "KAFKA_ENABLE_KRAFT": "false",
                **extra_bitnami_env,
            },
        )

        self.network = anonymous_network(docker_client)
        self.network.connect(self.zookeeper, aliases=["zk"])
        self.network.connect(self.broker, aliases=[self.static_broker_alias])
        super().__init__((self.zookeeper, self.broker), **kwargs)

    def connection_port(self):
        self.broker.reload()
        ports = get_ports(self.broker)
        return ports[self.outer_port]

    def _consumer(self, **kwargs) -> ContextManager[Any]:
        port = self.connection_port()
        if KafkaConsumer is not None:
            return closing(
                KafkaConsumer(
                    bootstrap_servers=[f"{DOCKER_EXPOSE_HOST}:{port}"], security_protocol="PLAINTEXT", **kwargs
                )
            )
        else:
            return closing(
                ConfluentConsumer(
                    {
                        "bootstrap.servers": f"{DOCKER_EXPOSE_HOST}:{port}",
                        "security.protocol": "PLAINTEXT",
                        "group.id": "yb-0",
                        **kwargs,
                    }
                )
            )

    def consumer(self, **kwargs) -> ContextManager[KafkaConsumer]:
        if KafkaConsumer is None:
            raise ImportError("kafka-python is not installed")
        return self._consumer(**kwargs)

    def producer(self, **kwargs) -> ContextManager[KafkaProducer]:
        if KafkaProducer is None:
            raise ImportError("kafka-python is not installed")
        port = self.connection_port()
        return cast(
            "ContextManager[KafkaProducer]",
            closing(
                KafkaProducer(
                    bootstrap_servers=[f"{DOCKER_EXPOSE_HOST}:{port}"], security_protocol="PLAINTEXT", **kwargs
                )
            ),
        )

    def start(self, retry_spec: RetrySpec | None = None):
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=20)
        with retry_spec.retry(self._consumer, (KafkaError, ConnectionError, ValueError, TypeError)):
            pass
        return self

    async def astart(self, retry_spec: RetrySpec | None = None) -> None:
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=20)
        with await retry_spec.aretry(self._consumer, (KafkaError, ConnectionError, ValueError, TypeError)):
            pass

    def stop(self, signal="SIGKILL"):
        # difference in default signal
        self.network.disconnect(self.broker)
        self.network.disconnect(self.zookeeper)
        self.network.remove()
        super().stop(signal)

    def connect(self, network, *, aliases=(), **kwargs):
        if not isinstance(aliases, list):
            aliases = list(aliases)
        aliases.append(self.static_broker_alias)
        kwargs["aliases"] = aliases
        return super().connect(network, **kwargs)

    @property
    def _single_endpoint(self):
        return self.broker



from collections import defaultdict
from collections.abc import Callable, Mapping
from contextlib import contextmanager
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer
from re import Pattern
from threading import Lock, Thread
from types import new_class
from typing import ClassVar, DefaultDict, NamedTuple, Union, cast
from urllib.parse import ParseResult, parse_qs, urlparse

import requests
from requests import ConnectionError, HTTPError

from yellowbox.retry import RetrySpec
from yellowbox.service import YellowService
from yellowbox.utils import docker_host_name

__all__ = ["HttpService", "RouterHTTPRequestHandler"]
SideEffectResponse = Union[bytes, str, int, "RouterHTTPRequestHandler"]
SideEffect = Union[Callable[["RouterHTTPRequestHandler"], SideEffectResponse], SideEffectResponse]


class RoutedHandler(NamedTuple):
    method: str
    name: str
    route: Pattern[str] | str
    callback: Callable[[RouterHTTPRequestHandler], None]

    def route_match(self, path: str):
        if isinstance(self.route, str):
            return self.route == path
        return self.route.fullmatch(path)


class RouterHTTPRequestHandler(BaseHTTPRequestHandler):
    """
    A BaseHTTPRequestHandler that allows adding and deleting routed handlers.
    Also contains some utility argument parsing.
    """

    _parse_url: ParseResult
    _body: bytes | None

    routes_by_method: ClassVar[DefaultDict[str, set[RoutedHandler]]]
    route_lock: ClassVar[Lock]

    def body(self) -> bytes | None:
        try:
            return self._body
        except AttributeError:
            raw_body_len = self.headers["Content-Length"]
            if raw_body_len is None:
                self._body = None
            else:
                length = int(raw_body_len)
                self._body = self.rfile.read(length)
            return self._body

    def path_params(self, **kwargs) -> Mapping[str, list[str]]:
        """
        Extract the path parameters from the query
        Args:
            **kwargs: forwarded to urllib.parse.parse_qs

        Returns:
            A mapping from parameter name to a list of values provided

        """
        parsed = self.parse_url()
        return parse_qs(parsed.query, **kwargs)

    def parse_url(self) -> ParseResult:
        try:
            return self._parse_url
        except AttributeError:
            self._parse_url = urlparse(self.path)
            return self._parse_url

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.routes_by_method = defaultdict(set)
        cls.route_lock = Lock()

    @classmethod
    def add_route(cls, handler: RoutedHandler):
        with cls.route_lock:
            cls.routes_by_method[handler.method].add(handler)

    @classmethod
    def del_route(cls, handler: RoutedHandler):
        with cls.route_lock:
            cls.routes_by_method[handler.method].remove(handler)

    def _do(self):
        """
        A generic handler for all http methods.
        Filters through all added routes and calls the appropriate callback, if one is found.
        """
        parsed = self.parse_url()

        candidates = tuple(self.routes_by_method.get(self.command, ()))
        matched_candidates = []
        for candidate in candidates:
            match = candidate.route_match(parsed.path)
            if match:
                matched_candidates.append((candidate, match))
        if not matched_candidates:
            self.send_error(404, "mock server matched no routes")
            self.end_headers()
        elif len(matched_candidates) > 1:
            self.send_error(
                500, "mock server matched multiple routes: " + ", ".join(c.name for c, _ in matched_candidates)
            )
            self.end_headers()
        else:
            ((routed, match),) = matched_candidates
            self.log_message(f"routed to {routed.name}")
            self.match = match
            routed.callback(self)

    def __getattr__(self, item: str):
        if item.startswith("do_"):
            return self._do
        raise AttributeError(item)


class HttpService(YellowService):
    """
    The HttpService class is used to mock http servers. Although it is a YellowService,
    it does not wrap a docker container, rather, it wraps a standard library HTTPServer, with a server
    thread running in the background. The server uses an internal router that maps paths to callbacks.

    Example usage:
    >>> with HttpService().start() as service:
    ...   @service.patch_route('GET', '/hello/world')
    ...   def hello_world(handler: RouterHTTPRequestHandler):
    ...      return "hi there"
    ...   # hello_world is now a context manager
    ...   with hello_world:
    ...      # within this scope, the path "/hello/world" will return a 200 response with the body "hi there"
    ...      assert requests.get(service.local_url+"/hello/world").text == "hi there"
    ...   # routes can also be set without a function
    ...   with service.patch_route('GET', '/meaning_of_life', '42'):
    ...      assert requests.get(service.local_url+"/meaning_of_life").content == b'42'
    """

    def __init__(self, host="0.0.0.0", port=0, name="anonymous_yellowbox_HTTPService"):
        self.router_cls = cast(
            type[RouterHTTPRequestHandler], new_class(name + "_RequestHandler", (RouterHTTPRequestHandler,))
        )
        self.server = HTTPServer((host, port), self.router_cls)
        self.server_thread = Thread(name=name + "_thread", target=self.server.serve_forever, daemon=True)

    @property
    def server_port(self):
        return self.server.server_port

    @property
    def local_url(self):
        return f"http://127.0.0.1:{self.server_port}"

    @property
    def container_url(self):
        return f"http://{docker_host_name}:{self.server_port}"

    @staticmethod
    def _to_callback(side_effect: SideEffect):
        def _respond(handler: RouterHTTPRequestHandler, response: SideEffectResponse):
            if response is handler:
                return  # Assuming the user already handled the response
            if isinstance(response, int):
                handler.send_error(response)
                handler.end_headers()
                return
            handler.send_response(200)
            handler.end_headers()
            if isinstance(response, str):
                response = bytes(response, "ascii")
            if isinstance(response, bytes):
                handler.wfile.write(response)
            else:
                raise TypeError(
                    f"got response of type {type(response)}, type must be RouterHTTPRequestHandler, int, str or bytes"
                )

        def callback(handler: RouterHTTPRequestHandler):
            if callable(side_effect):
                result = side_effect(handler)
                _respond(handler, result)
            else:
                _respond(handler, cast(SideEffectResponse, side_effect))

        return callback

    def patch_route(
        self,
        method,
        route: str | Pattern[str],
        side_effect: SideEffect = ...,  # type: ignore[assignment]
        name: str | None = None,
    ):
        """
        Create a context manager that temporarily adds a route handler to the service.

        Args:
            method: The request method to add the route to.
            route: The route to attach the side effect to, all routes must begin with a slash "/".
                Alternatively, The route may be a regex pattern, in which case the request path must fully match it,
                the match object is then stored in RouterHTTPRequestHandler.match, to be used by a side-effect callable.
            side_effect: The callback or result to return for the route. Accepts any of the following types:
                * int: to return the value as the HTTP status code, without a body.
                * bytes: to return 200, with the value as the response body.
                * str: invalid if the value is non-ascii, return 200 with the value, translated to bytes, as
                 the response body.
                * callable: Must accept a RouterHTTPRequestHandler. May return any of the above types, or
                  RouterHTTPRequestHandler to handle the response directly with the RouterHTTPRequestHandler.
            name:
                An optional name for the routed handler, to be used while logging. If missing, a suitable name is
                extracted from the side effect or route.

        Returns:
            A context manager that will enable the route upon entry and disable it upon exit.

        Notes:
            This method can be used as a decorator by not specifying `side_effect`
        """
        if side_effect is ...:
            return partial(self.patch_route, method, route, name=name)

        @contextmanager
        def _helper():
            nonlocal name
            callback = self._to_callback(side_effect)
            name = name or getattr(side_effect, "__name__", None) or str(route)
            handler = RoutedHandler(method, name, route, callback)
            self.router_cls.add_route(handler)
            try:
                yield handler
            finally:
                self.router_cls.del_route(handler)

        return _helper()

    def start(self, retry_spec: RetrySpec | None = None):
        with self.patch_route("GET", "/health", 200):
            self.server_thread.start()
            retry_spec = retry_spec or RetrySpec(attempts=10)
            retry_spec.retry(
                lambda: requests.get(self.local_url + "/health").raise_for_status(), (ConnectionError, HTTPError)
            )
        return super().start()

    def stop(self):
        self.server.shutdown()
        self.server_thread.join()

    def is_alive(self):
        return self.server_thread.is_alive()



"""
Azure Blob Storage module, for creating container, uploading files to it and downloading files.
"""

from collections.abc import Sequence
from typing import Any

from docker import DockerClient
from docker.models.networks import Network

from yellowbox.containers import create_and_pull_with_defaults, get_ports, short_id
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, RunMixin, SingleContainerService

__all__ = ["AzuriteService", "BlobStorageService"]

from yellowbox.utils import DOCKER_EXPOSE_HOST, docker_host_name

BLOB_STORAGE_DEFAULT_PORT = 10000
DEFAULT_ACCOUNT_KEY = "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw=="
DEFAULT_ACCOUNT_NAME = "devstoreaccount1"


class _ResourceNotReadyError(Exception):
    pass


class AzuriteService(SingleContainerService, RunMixin, AsyncRunMixin):
    """
    Starts Azurite, Azure's storage emulator.
    Provides helper functions for preparing the instance for testing.
    TODO: Make account name and key configurable.
    """

    account_name = DEFAULT_ACCOUNT_NAME
    account_key = DEFAULT_ACCOUNT_KEY

    def __init__(
        self,
        docker_client: DockerClient,
        image: str = "mcr.microsoft.com/azure-storage/azurite:latest",
        *,
        skip_api_version_check: bool = True,
        container_create_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ):
        command = "azurite-blob --blobHost 0.0.0.0"
        if skip_api_version_check:
            command += " --skipApiVersionCheck"

        container = create_and_pull_with_defaults(
            docker_client,
            image,
            command,
            _kwargs=container_create_kwargs,
            publish_all_ports=True,
        )
        super().__init__(container, **kwargs)

    def stop(self, signal: str | int = "SIGKILL"):
        """
        We override to change the signal.
        """
        super().stop(signal)

    def client_port(self):
        return get_ports(self.container)[BLOB_STORAGE_DEFAULT_PORT]

    @property
    def connection_string(self):
        """Connection string to connect from host to container"""
        return (
            f"DefaultEndpointsProtocol=http;"
            f"AccountName={self.account_name};"
            f"AccountKey={self.account_key};"
            f"BlobEndpoint={self.endpoint_url};"
        )

    @property
    def container_connection_string(self):
        """Connection string to connect across containers in the same network"""
        return (
            f"DefaultEndpointsProtocol=http;"
            f"AccountName={self.account_name};"
            f"AccountKey={self.account_key};"
            f"BlobEndpoint={self.container_endpoint_url};"
        )

    @property
    def host_connection_string(self):
        """Connection string to connect across containers through the docker host"""
        return (
            f"DefaultEndpointsProtocol=http;"
            f"AccountName={self.account_name};"
            f"AccountKey={self.account_key};"
            f"BlobEndpoint={self.host_endpoint_url};"
        )

    @property
    def endpoint_url(self):
        """URL for the endpoint from docker host"""
        return f"http://{DOCKER_EXPOSE_HOST}:{self.client_port()}/{self.account_name}"

    @property
    def container_endpoint_url(self):
        """URL for the endpoint from another container over a common network"""
        return f"http://{short_id(self.container)}:{BLOB_STORAGE_DEFAULT_PORT}/{self.account_name}"

    @property
    def host_endpoint_url(self):
        """URL for the endpoint from another container through the docker host"""
        return f"http://{docker_host_name}:{self.client_port()}/{self.account_name}"

    @property
    def account_credentials(self):
        """Azure credentials dict to connect to the service"""
        return {"account_name": self.account_name, "account_key": self.account_key}

    def _check_ready(self):
        if b"Azurite Blob service successfully listens on" not in self.container.logs():
            raise _ResourceNotReadyError

    def start(self, retry_spec: RetrySpec | None = None):
        super().start()

        retry_spec = retry_spec or RetrySpec(attempts=10)

        retry_spec.retry(self._check_ready, _ResourceNotReadyError)
        return self

    async def astart(self, retry_spec: RetrySpec | None = None):
        super().start()

        retry_spec = retry_spec or RetrySpec(attempts=10)

        await retry_spec.aretry(self._check_ready, _ResourceNotReadyError)
        return self

    def connect(self, network: Network, aliases: list[str] | None = None, **kwargs) -> Sequence[str]:
        # Make sure the id is in the aliases list. Needed for the container
        # connection string.
        if aliases is not None:
            aliases.append(short_id(self.container))
        return super().connect(network, aliases=aliases, **kwargs)


# legacy alias
BlobStorageService = AzuriteService



from collections.abc import Iterable
from contextlib import contextmanager
from os import environ, getenv
from typing import Any
from warnings import warn

import requests
from docker import DockerClient

from yellowbox.containers import create_and_pull_with_defaults, get_ports
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, RunMixin, SingleContainerService
from yellowbox.utils import DOCKER_EXPOSE_HOST, docker_host_name

FAKE_GCS_DEFAULT_PORT = 4443


class FakeGoogleCloudStorage(SingleContainerService, RunMixin, AsyncRunMixin):
    def __init__(
        self,
        docker_client: DockerClient,
        image: str = "fsouza/fake-gcs-server:1.42",
        scheme: str = "https",
        command: str = "",
        *,
        container_create_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ):
        # note that fake-gcs-server 1.43.0 has a bug https://github.com/fsouza/fake-gcs-server/issues/1034
        command = f"-scheme {scheme} {command}"
        self.scheme = scheme
        container = create_and_pull_with_defaults(
            docker_client, image, command, _kwargs=container_create_kwargs, publish_all_ports=True
        )
        super().__init__(container, **kwargs)

    def client_port(self):
        return get_ports(self.container)[FAKE_GCS_DEFAULT_PORT]

    def local_url(self, scheme: str | None = ...):  # type: ignore[assignment]
        ret = f"{DOCKER_EXPOSE_HOST}:{self.client_port()}"
        if scheme is ...:
            scheme = self.scheme
        if scheme:
            ret = f"{self.scheme}://{ret}"
        return ret

    def container_url(self, hostname: str, scheme: str | None = ...):  # type: ignore[assignment]
        ret = f"{hostname}:{FAKE_GCS_DEFAULT_PORT}"
        if scheme is ...:
            scheme = self.scheme
        if scheme:
            ret = f"{self.scheme}://{ret}"
        return ret

    def host_url(self, scheme: str | None = ...):  # type: ignore[assignment]
        ret = f"{docker_host_name}:{self.client_port()}"
        if scheme is ...:
            scheme = self.scheme
        if scheme:
            ret = f"{self.scheme}://{ret}"
        return ret

    def start(self, retry_spec: RetrySpec | None = None):
        super().start()
        url = self.local_url() + "/storage/v1/b"
        retry_spec = retry_spec or RetrySpec(attempts=15)
        retry_spec.retry(
            lambda: requests.get(url, verify=False).raise_for_status(), requests.exceptions.RequestException
        )
        return self

    async def astart(self, retry_spec: RetrySpec | None = None):
        super().start()
        url = self.local_url() + "/storage/v1/b"
        retry_spec = retry_spec or RetrySpec(attempts=15)
        await retry_spec.aretry(
            lambda: requests.get(url, verify=False).raise_for_status(), requests.exceptions.RequestException
        )

    @contextmanager
    def patch_gcloud_aio(self):
        from gcloud.aio.storage import __version__ as gcloud_aio_version  # noqa: PLC0415

        if not gcloud_aio_version.startswith("7."):
            # for newer gcloud_aio, we can just adjust the environment
            warn(
                "newer gcloud versions should be patched directly by setting the environment variable "
                "STORAGE_EMULATOR_HOST to service.local_url()",
                stacklevel=1,
            )
            prev_env = getenv("STORAGE_EMULATOR_HOST")
            environ["STORAGE_EMULATOR_HOST"] = self.local_url(None)
            yield
            if prev_env is None:
                del environ["STORAGE_EMULATOR_HOST"]
            else:
                environ["STORAGE_EMULATOR_HOST"] = prev_env
            return
        import gcloud.aio.storage.storage as gcloud_module  # noqa: PLC0415

        previous_state = (
            gcloud_module.API_ROOT,
            gcloud_module.API_ROOT_UPLOAD,
            gcloud_module.VERIFY_SSL,
            gcloud_module.STORAGE_EMULATOR_HOST,
        )
        (
            gcloud_module.API_ROOT,
            gcloud_module.API_ROOT_UPLOAD,
            gcloud_module.VERIFY_SSL,
            gcloud_module.STORAGE_EMULATOR_HOST,
        ) = (
            self.local_url() + "/storage/v1/b",
            self.local_url() + "/upload/storage/v1/b",
            False,
            self.local_url(scheme=""),
        )
        yield
        (
            gcloud_module.API_ROOT,
            gcloud_module.API_ROOT_UPLOAD,
            gcloud_module.VERIFY_SSL,
            gcloud_module.STORAGE_EMULATOR_HOST,
        ) = previous_state

    def create_bucket(self, bucket_name: str) -> dict[str, Any]:
        url = self.local_url()

        resp = requests.post(url + "/storage/v1/b", json={"name": bucket_name}, verify=False)
        resp.raise_for_status()

        return resp.json()

    def clear_bucket(self, bucket_name: str, prefix: str | None = None) -> Iterable[str]:
        url = self.local_url()
        params = {}
        if prefix:
            params["prefix"] = prefix
        page_token = None
        ret = []
        while True:
            if page_token:
                params["pageToken"] = page_token
            resp = requests.get(url + f"/storage/v1/b/{bucket_name}/o", params=params, verify=False)
            resp.raise_for_status()
            data = resp.json()
            for item in data["items"]:
                ret.append(item["name"])
                requests.delete(url + f"/storage/v1/b/{bucket_name}/o/{item['name']}", verify=False).raise_for_status()
            if "nextPageToken" in data:
                page_token = data["next_page_token"]
            else:
                break
        return ret

    def delete_bucket(self, bucket_name: str, force: bool = False, missing_ok: bool = False):
        url = self.local_url()

        try:
            if force:
                # we need to delete all the objects in the bucket
                self.clear_bucket(bucket_name)

            resp = requests.delete(url + f"/storage/v1/b/{bucket_name}", verify=False)
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404 or not missing_ok:
                raise
